@@ -39,12 +39,7 @@
  *                                                                           *
  * ========================================================================= */
 
-/*===========================================================================*\
- *                                                                           *             
- *   $Revision$                                                         *
- *   $Date$                   *
- *                                                                           *
-\*===========================================================================*/
+
 
 //== IMPLEMENTATION ==========================================================
 #include <OpenMesh/Core/Mesh/PolyConnectivity.hh>
@@ -59,8 +54,7 @@ const PolyConnectivity::FaceHandle      PolyConnectivity::InvalidFaceHandle;
 
 //-----------------------------------------------------------------------------
 
-PolyConnectivity::HalfedgeHandle
-PolyConnectivity::find_halfedge(VertexHandle _start_vh, VertexHandle _end_vh ) const
+SmartHalfedgeHandle PolyConnectivity::find_halfedge(VertexHandle _start_vh, VertexHandle _end_vh ) const
 {
   assert(_start_vh.is_valid() && _end_vh.is_valid());
 
@@ -68,7 +62,7 @@ PolyConnectivity::find_halfedge(VertexHandle _start_vh, VertexHandle _end_vh ) c
     if (to_vertex_handle(*voh_it) == _end_vh)
       return *voh_it;
 
-  return InvalidHalfedgeHandle;
+  return make_smart(InvalidHalfedgeHandle, this);
 }
 
 
@@ -104,6 +98,7 @@ bool PolyConnectivity::is_manifold(VertexHandle _vh) const
 }
 
 //-----------------------------------------------------------------------------
+
 void PolyConnectivity::adjust_outgoing_halfedge(VertexHandle _vh)
 {
   for (ConstVertexOHalfedgeIter vh_it=cvoh_iter(_vh); vh_it.is_valid(); ++vh_it)
@@ -118,7 +113,7 @@ void PolyConnectivity::adjust_outgoing_halfedge(VertexHandle _vh)
 
 //-----------------------------------------------------------------------------
 
-PolyConnectivity::FaceHandle
+SmartFaceHandle
 PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size)
 {
   VertexHandle                   vh;
@@ -135,7 +130,8 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
     edgeData_.resize(n);
     next_cache_.resize(6*n);
   }
-  next_cache_count_ = 0;
+
+  size_t next_cache_count = 0;
 
   // don't allow degenerated faces
   assert (n > 2);
@@ -146,7 +142,7 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
     if ( !is_boundary(_vertex_handles[i]) )
     {
       omerr() << "PolyMeshT::add_face: complex vertex\n";
-      return InvalidFaceHandle;
+      return make_smart(InvalidFaceHandle, this);
     }
 
     // Initialise edge attributes
@@ -158,7 +154,7 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
     if (!edgeData_[i].is_new && !is_boundary(edgeData_[i].halfedge_handle))
     {
       omerr() << "PolyMeshT::add_face: complex edge\n";
-      return InvalidFaceHandle;
+      return make_smart(InvalidFaceHandle, this);
     }
   }
 
@@ -190,7 +186,7 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
         if (boundary_prev == inner_prev)
         {
           omerr() << "PolyMeshT::add_face: patch re-linking failed\n";
-          return InvalidFaceHandle;
+          return make_smart(InvalidFaceHandle, this);
         }
 
         assert(is_boundary(boundary_prev));
@@ -208,9 +204,9 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
         assert(inner_next.is_valid());
 
         // relink
-        next_cache_[next_cache_count_++] = std::make_pair(boundary_prev, patch_start);
-        next_cache_[next_cache_count_++] = std::make_pair(patch_end, boundary_next);
-        next_cache_[next_cache_count_++] = std::make_pair(inner_prev, inner_next);
+        next_cache_[next_cache_count++] = std::make_pair(boundary_prev, patch_start);
+        next_cache_[next_cache_count++] = std::make_pair(patch_end, boundary_next);
+        next_cache_[next_cache_count++] = std::make_pair(inner_prev, inner_next);
       }
     }
   }
@@ -252,14 +248,14 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
         case 1: // prev is new, next is old
           boundary_prev = prev_halfedge_handle(inner_next);
           assert(boundary_prev.is_valid());
-          next_cache_[next_cache_count_++] = std::make_pair(boundary_prev, outer_next);
+          next_cache_[next_cache_count++] = std::make_pair(boundary_prev, outer_next);
           set_halfedge_handle(vh, outer_next);
           break;
 
         case 2: // next is new, prev is old
           boundary_next = next_halfedge_handle(inner_prev);
           assert(boundary_next.is_valid());
-          next_cache_[next_cache_count_++] = std::make_pair(outer_prev, boundary_next);
+          next_cache_[next_cache_count++] = std::make_pair(outer_prev, boundary_next);
           set_halfedge_handle(vh, boundary_next);
           break;
 
@@ -267,7 +263,7 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
           if (!halfedge_handle(vh).is_valid())
           {
             set_halfedge_handle(vh, outer_next);
-            next_cache_[next_cache_count_++] = std::make_pair(outer_prev, outer_next);
+            next_cache_[next_cache_count++] = std::make_pair(outer_prev, outer_next);
           }
           else
           {
@@ -275,14 +271,14 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
             boundary_prev = prev_halfedge_handle(boundary_next);
             assert(boundary_prev.is_valid());
             assert(boundary_next.is_valid());
-            next_cache_[next_cache_count_++] = std::make_pair(boundary_prev, outer_next);
-            next_cache_[next_cache_count_++] = std::make_pair(outer_prev, boundary_next);
+            next_cache_[next_cache_count++] = std::make_pair(boundary_prev, outer_next);
+            next_cache_[next_cache_count++] = std::make_pair(outer_prev, boundary_next);
           }
           break;
       }
 
       // set inner link
-      next_cache_[next_cache_count_++] = std::make_pair(inner_prev, inner_next);
+      next_cache_[next_cache_count++] = std::make_pair(inner_prev, inner_next);
     }
     else edgeData_[ii].needs_adjust = (halfedge_handle(vh) == inner_next);
 
@@ -292,7 +288,7 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
   }
 
   // process next halfedge cache
-  for (i = 0; i < next_cache_count_; ++i)
+  for (i = 0; i < next_cache_count; ++i)
     set_next_halfedge_handle(next_cache_[i].first, next_cache_[i].second);
 
 
@@ -301,12 +297,12 @@ PolyConnectivity::add_face(const VertexHandle* _vertex_handles, size_t _vhs_size
     if (edgeData_[i].needs_adjust)
       adjust_outgoing_halfedge(_vertex_handles[i]);
 
-  return fh;
+  return make_smart(fh, this);
 }
 
 //-----------------------------------------------------------------------------
 
-FaceHandle PolyConnectivity::add_face(VertexHandle _vh0, VertexHandle _vh1, VertexHandle _vh2, VertexHandle _vh3)
+SmartFaceHandle PolyConnectivity::add_face(VertexHandle _vh0, VertexHandle _vh1, VertexHandle _vh2, VertexHandle _vh3)
 {
   VertexHandle vhs[4] = { _vh0, _vh1, _vh2, _vh3 };
   return add_face(vhs, 4);
@@ -314,7 +310,7 @@ FaceHandle PolyConnectivity::add_face(VertexHandle _vh0, VertexHandle _vh1, Vert
 
 //-----------------------------------------------------------------------------
 
-FaceHandle PolyConnectivity::add_face(VertexHandle _vh0, VertexHandle _vh1, VertexHandle _vh2)
+SmartFaceHandle PolyConnectivity::add_face(VertexHandle _vh0, VertexHandle _vh1, VertexHandle _vh2)
 {
   VertexHandle vhs[3] = { _vh0, _vh1, _vh2 };
   return add_face(vhs, 3);
@@ -322,9 +318,17 @@ FaceHandle PolyConnectivity::add_face(VertexHandle _vh0, VertexHandle _vh1, Vert
 
 //-----------------------------------------------------------------------------
 
-FaceHandle PolyConnectivity::add_face(const std::vector<VertexHandle>& _vhandles)
+SmartFaceHandle PolyConnectivity::add_face(const std::vector<VertexHandle>& _vhandles)
 { return add_face(&_vhandles.front(), _vhandles.size()); }
 
+//-----------------------------------------------------------------------------
+
+SmartFaceHandle PolyConnectivity::add_face(const std::vector<SmartVertexHandle>& _vhandles)
+{
+  std::vector<VertexHandle> vhandles(_vhandles.begin(), _vhandles.end());
+  return add_face(&vhandles.front(), vhandles.size());
+}
+
 
 //-----------------------------------------------------------------------------
 bool PolyConnectivity::is_collapse_ok(HalfedgeHandle v0v1)
@@ -363,14 +367,16 @@ bool PolyConnectivity::is_collapse_ok(HalfedgeHandle v0v1)
 
   //the edges v1-vl and vl-v0 must not be both boundary edges
   //this test makes only sense in a polymesh if the side face is a triangle
+  VertexHandle vl;
   if (!is_boundary(v0v1))
   {
     if (v0v1_triangle)
     {
-      //VertexHandle vl = to_vertex_handle(next_halfedge_handle(v0v1));
-
       HalfedgeHandle h1 = next_halfedge_handle(v0v1);
       HalfedgeHandle h2 = next_halfedge_handle(h1);
+      
+      vl = to_vertex_handle(h1);
+      
       if (is_boundary(opposite_halfedge_handle(h1)) && is_boundary(opposite_halfedge_handle(h2)))
         return false;
     }
@@ -378,19 +384,24 @@ bool PolyConnectivity::is_collapse_ok(HalfedgeHandle v0v1)
 
   //the edges v0-vr and vr-v1 must not be both boundary edges
   //this test makes only sense in a polymesh if the side face is a triangle
+  VertexHandle vr;
   if (!is_boundary(v1v0))
   {
     if (v1v0_triangle)
     {
-      //VertexHandle vr = to_vertex_handle(next_halfedge_handle(v1v0));
-
       HalfedgeHandle h1 = next_halfedge_handle(v1v0);
       HalfedgeHandle h2 = next_halfedge_handle(h1);
+      
+      vr = to_vertex_handle(h1);
+      
       if (is_boundary(opposite_halfedge_handle(h1)) && is_boundary(opposite_halfedge_handle(h2)))
         return false;
     }
   }
 
+  // if vl and vr are equal and valid (e.g. triangle case) -> fail
+  if ( vl.is_valid() && (vl == vr)) return false;
+
   // edge between two boundary vertices should be a boundary edge
   if ( is_boundary(v0) && is_boundary(v1) && !is_boundary(v0v1) && !is_boundary(v1v0))
     return false;
@@ -620,101 +631,6 @@ void PolyConnectivity::delete_face(FaceHandle _fh, bool _delete_isolated_vertice
     adjust_outgoing_halfedge(*v_it);
 }
 
-//-----------------------------------------------------------------------------
-PolyConnectivity::VertexIter PolyConnectivity::vertices_begin()
-{
-  return VertexIter(*this, VertexHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstVertexIter PolyConnectivity::vertices_begin() const
-{
-  return ConstVertexIter(*this, VertexHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::VertexIter PolyConnectivity::vertices_end()
-{
-  return VertexIter(*this, VertexHandle( int(n_vertices() ) ));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstVertexIter PolyConnectivity::vertices_end() const
-{
-  return ConstVertexIter(*this, VertexHandle( int(n_vertices()) ));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::HalfedgeIter PolyConnectivity::halfedges_begin()
-{
-  return HalfedgeIter(*this, HalfedgeHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstHalfedgeIter PolyConnectivity::halfedges_begin() const
-{
-  return ConstHalfedgeIter(*this, HalfedgeHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::HalfedgeIter PolyConnectivity::halfedges_end()
-{
-  return HalfedgeIter(*this, HalfedgeHandle(int(n_halfedges())));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstHalfedgeIter PolyConnectivity::halfedges_end() const
-{
-  return ConstHalfedgeIter(*this, HalfedgeHandle(int(n_halfedges())));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::EdgeIter PolyConnectivity::edges_begin()
-{
-  return EdgeIter(*this, EdgeHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstEdgeIter PolyConnectivity::edges_begin() const
-{
-  return ConstEdgeIter(*this, EdgeHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::EdgeIter PolyConnectivity::edges_end()
-{
-  return EdgeIter(*this, EdgeHandle(int(n_edges())));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstEdgeIter PolyConnectivity::edges_end() const
-{
-  return ConstEdgeIter(*this, EdgeHandle(int(n_edges())));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::FaceIter PolyConnectivity::faces_begin()
-{
-  return FaceIter(*this, FaceHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstFaceIter PolyConnectivity::faces_begin() const
-{
-  return ConstFaceIter(*this, FaceHandle(0));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::FaceIter PolyConnectivity::faces_end()
-{
-  return FaceIter(*this, FaceHandle(int(n_faces())));
-}
-
-//-----------------------------------------------------------------------------
-PolyConnectivity::ConstFaceIter PolyConnectivity::faces_end() const
-{
-  return ConstFaceIter(*this, FaceHandle(int(n_faces())));
-}
 
 //-----------------------------------------------------------------------------
 void PolyConnectivity::collapse(HalfedgeHandle _hh)
@@ -776,6 +692,11 @@ void PolyConnectivity::collapse_edge(HalfedgeHandle _hh)
   // delete stuff
   status(edge_handle(h)).set_deleted(true);
   status(vo).set_deleted(true);
+  if (has_halfedge_status())
+  {
+    status(h).set_deleted(true);
+    status(o).set_deleted(true);
+  }
 }
 
 //-----------------------------------------------------------------------------
@@ -826,6 +747,11 @@ void PolyConnectivity::collapse_loop(HalfedgeHandle _hh)
     status(fh).set_deleted(true);
   }
   status(edge_handle(h0)).set_deleted(true);
+  if (has_halfedge_status())
+  {
+    status(h0).set_deleted(true);
+    status(o0).set_deleted(true);
+  }
 }
 
 //-----------------------------------------------------------------------------
@@ -1031,6 +957,7 @@ void PolyConnectivity::triangulate(FaceHandle _fh)
 
   HalfedgeHandle base_heh(halfedge_handle(_fh));
   VertexHandle start_vh = from_vertex_handle(base_heh);
+  HalfedgeHandle prev_heh(prev_halfedge_handle(base_heh));
   HalfedgeHandle next_heh(next_halfedge_handle(base_heh));
 
   while (to_vertex_handle(next_halfedge_handle(next_heh)) != start_vh)
@@ -1050,6 +977,10 @@ void PolyConnectivity::triangulate(FaceHandle _fh)
     set_face_handle(next_heh, new_fh);
     set_face_handle(new_heh,  new_fh);
 
+    copy_all_properties(prev_heh, new_heh, true);
+    copy_all_properties(prev_heh, opposite_halfedge_handle(new_heh), true);
+    copy_all_properties(_fh, new_fh, true);
+
     base_heh = opposite_halfedge_handle(new_heh);
     next_heh = next_next_heh;
   }
@@ -1124,7 +1055,7 @@ void PolyConnectivity::split_copy(FaceHandle fh, VertexHandle vh) {
 
   // Copy the property of the original face to all new faces
   for(VertexFaceIter vf_it =  vf_iter(vh); vf_it.is_valid(); ++vf_it)
-    copy_all_properties(fh, *vf_it);
+    copy_all_properties(fh, *vf_it, true);
 }
 
 //-----------------------------------------------------------------------------
@@ -1220,7 +1151,7 @@ void PolyConnectivity::split_edge_copy(EdgeHandle _eh, VertexHandle _vh)
   EdgeHandle eh0 = edge_handle( next_halfedge_handle( halfedge_handle(_eh, 1) ) );
 
   // Copy the property from the original to the new edge
-  copy_all_properties(_eh, eh0);
+  copy_all_properties(_eh, eh0, true);
 }
 
 } // namespace OpenMesh
